import os
import collections
import related

from pysam import VariantFile

from .calculate_vaf import find as find_vaf_map
from .bed_filter import ValidChromosomeFilter


def iterate_vcf_files(
    file_paths,
    include_vaf=True,
    bed_filter=None,
    build=None,
    keep_ref_unknown_calls=False,
):
    assert not isinstance(file_paths, str), "Requires list or iterator of str."
    for file_path in file_paths:
        for call in iterate_vcf_calls(
            file_path,
            include_vaf=include_vaf,
            bed_filter=bed_filter,
            build=build,
            keep_ref_unknown_calls=keep_ref_unknown_calls,
        ):
            yield call


def iterate_vcf_calls(
    file_path,
    include_vaf=True,
    bed_filter=None,
    build=None,
    keep_ref_unknown_calls=False,
):
    vcf_iter = create_vcf_iter(file_path)

    header = create_header_dict(next(vcf_iter), include_vaf, file_path)
    yield header

    vaf_map = find_vaf_map(header.get("types").keys()) if include_vaf else None

    bed_filter = bed_filter or ValidChromosomeFilter()

    for record in filter(bed_filter, vcf_iter):
        samples = get_samples(record)

        for sample in samples:
            # if GT not specified, then assume single alt
            allele_indices, zyg, sample_is_ref_unknown = \
                get_allele_indices_and_zyg(sample)

            for allele_index in allele_indices:
                alt = get_alt(record, allele_index)
                info = capture_info(header, record, sample, allele_index)

                # VCF 4.2: '*' indicates allele missing due to upstream del
                if alt in ["*", "."]:
                    yield create_skipped_call_dict(record, alt, info.get("GT"))
                elif keep_ref_unknown_calls or not sample_is_ref_unknown:
                    vaf_dict = (
                        vaf_map(info, allele_index, alt) if vaf_map else {}
                    )
                    yield create_call_dict(
                        record, sample, alt, zyg, info, vaf_dict, build
                    )


# private methods


@related.immutable()
class SampleSubstitute(object):
    name = related.StringField(default="sample")
    allele_indices = related.SequenceField(int, default=[1])
    alleles = related.SequenceField(str, default=["?"])
    phased = related.BooleanField(default=False)

    @staticmethod
    def iteritems():
        return []

    @property
    def allele_indices(self):
        return range(len(self.alleles))


substitute_sample = SampleSubstitute()


def get_alt(record, allele_index):
    try:
        alt = record.alleles[allele_index]
    except IndexError:
        alt = "."
    return alt


def get_samples(record):
    samples = list(record.samples.itervalues())
    if not samples:
        samples = [SampleSubstitute(alleles=record.alleles)]
    return samples


def create_vcf_iter(file_path):
    vcf = VariantFile(file_path)
    yield vcf.header

    for vcf_record in vcf.fetch():
        yield vcf_record


def create_header_dict(header, include_vaf, file_path):
    return dict(
        __type__="HEADER",
        __child__="CALL",
        __record__=header,
        file_path=os.path.realpath(file_path),
        config=dict(include_vaf=include_vaf),
        types=capture_solr_types(header),
        info=create_meta_dict(header.info.items()),
        formats=create_meta_dict(header.formats.items()),
    )


def iter_attrs(obj, include_private=False, type_filter=None):
    for attr_name in dir(obj):
        if include_private or not attr_name.startswith("_"):
            attr_value = getattr(obj, attr_name)
            if not type_filter or isinstance(attr_value, type_filter):
                yield (attr_name, attr_value)


def create_meta_dict(items, meta_types=(int, str, bool, float)):
    meta = {}
    for (key, value) in items:
        meta[key] = dict(iter_attrs(value, type_filter=meta_types))
    return meta


def map_chr(chromosome):
    if chromosome == "M":
        return "MT"
    return chromosome


def create_call_dict(record, sample, alt, zyg, info, vaf_dict, build):
    return dict(
        __type__="CALL",
        __record__=record,
        chr=map_chr(record.chrom.replace("chr", "")),
        start=record.pos,
        end=record.stop,
        ref=record.ref,
        alt=alt,
        sample_name=sample.name,
        zyg=zyg,
        is_phased=sample.phased,
        quality=record.qual,
        rs_id=record.id,
        filters=list(record.filter),
        info=info,
        build=build,
        **vaf_dict,
    )


def create_skipped_call_dict(record, alt, gt):
    return dict(
        __type__="SKIPPED_CALL",
        chr=map_chr(record.chrom.replace("chr", "")),
        start=record.pos,
        end=record.stop,
        ref=record.ref,
        alt=alt,
        GT=gt,
    )


def get_allele_indices_and_zyg(sample):
    allele_indices = sample.allele_indices or [1]
    zyg = determine_zyg(allele_indices)
    allele_indices = sorted(set(filter(None, allele_indices)))

    # keep all the 0/0 and ./. calls (will filter them out later)
    if not allele_indices:
        return [1], zyg, True

    return allele_indices, zyg, False


def capture_value(value, meta, allele_index):
    if meta:
        number = meta.get("number")

        # todo: additional thought required here and impact to calc vaf mappers
        # todo: for example, "R" might best return 2 values: REF + *THIS* ALT
        if number == "A":
            idx = allele_index - 1
            value = ensure_collection(value)
            value = value[idx] if idx < len(value) else value
    return value


def capture_by_section(iteritems, meta, allele_index):
    info = {}

    try:
        for (name, value) in iteritems:
            field_meta = meta.get(name)
            value = capture_value(value, field_meta, allele_index)
            info[name] = value

    except ValueError:
        # Note: https://github.com/pysam-developers/pysam/issues/593
        pass

    return info


def capture_info(header, record, sample, allele_index):

    info = capture_by_section(
        record.info.iteritems(), header.get("info"), allele_index
    )

    info.update(
        capture_by_section(
            sample.iteritems(), header.get("formats"), allele_index
        )
    )

    info["GT"] = format_GT(info, sample)

    return info


def capture_solr_types(header):
    def to_solr_type(meta):
        multi = "" if meta.number in {0, 1, "A"} else "m"
        meta_type = TYPES.get(meta.type, "string")
        return multi + meta_type

    solr_t = dict([(m.name, to_solr_type(m)) for m in header.info.values()])
    solr_t.update([(m.name, to_solr_type(m)) for m in header.formats.values()])
    return solr_t


def ensure_collection(x):
    """ Ensures the parameter provided is an iterable """

    if x is None:
        x = []

    elif isinstance(x, collections.Iterator):
        x = list(x)

    elif not isinstance(x, (list, tuple, set)):
        x = [x]

    return x


def determine_zyg(allele_indicies):
    if len(allele_indicies) == 1:
        zyg = "hemizygous"
    elif None in allele_indicies:
        zyg = "unknown"

    elif allele_indicies[0] != allele_indicies[1]:
        zyg = "heterozygous"

    elif allele_indicies[0] == allele_indicies[1]:
        zyg = "homozygous"

    else:
        zyg = "unknown"

    return zyg


def format_GT(info, sample):
    gt = info.get("GT")
    if gt:
        sep = "|" if sample.phased else "/"
        gt = map(lambda g: "." if g is None else str(g), gt)
        return sep.join(gt)


# constants

TYPES = dict(Integer="int", Float="float", Flag="boolean")



from cytoolz.curried import curry
from gosdk.logger import get_logger
from .base import FileSink


@curry
class VcfFileSink(FileSink):
    def __init__(self, filename):
        super().__init__(filename, insert_newlines=False)
        self.records_seen = set()
        self.count = 0

    def convert(self, unit):
        self.count += 1
        record = unit.get("__record__")

        # if record is None, return None after warning
        if record is None:
            get_logger().warn(f"Missing __record__: {unit}")
            return

        # return header as string
        if self.count == 1:
            return str(record)

        # check and add unique records
        record_id = f"{record.chrom}:{record.pos}"
        if record_id not in self.records_seen:
            self.records_seen.add(record_id)
            return str(record)



from enum import Enum

__TYPE__ = "__type__"
__CHILD__ = "__child__"


class DocType(Enum):
    ANNOTATED_CALL = "ANNOTATED_CALL"
    ANNOTATED_VARIANT = "ANNOTATED_VARIANT"
    ANNOTATED_TSV = "ANNOTATED_TSV"
    CALL = "CALL"
    SKIPPED_CALL = "SKIPPED_CALL"
    FEATURE = "FEATURE"
    GENE = "GENE"
    HEADER = "HEADER"
    VARIANT = "VARIANT"
    TRIAL = "TRIAL"
    CONTENT = "CONTENT"
    THERAPY = "THERAPY"
    TUXEDO = "TUXEDO"
    ALTERATION = "ALTERATION"
    MARKER = "MARKER"

    TSV = "TSV"
    AGGREGATE = "AGGREGATE"
    MAF = "MAF"

    UNIPROT_RECORD = "UNIPROT_RECORD"

    MITOMAP_RECORD = "MITOMAP_RECORD"

    # Annotation Data Sets
    GNOMAD = "GNOMAD"
    HGMD = "HGMD"
    BAYLOR_WGL = "BAYLOR_WGL"
    DBNSFP = "DBNSFP"
    CLINVAR = "CLINVAR"
    THOUSANDGENOMES = "THOUSANDGENOMES"
    DBSNP = "DBSNP"
    EXAC = "EXAC"
    SPLICING = "SPLICING"
    EVS = "EVS"
    COSMIC = "COSMIC"
    BAYLOR_DECIPHER = "BAYLOR_DECIPHER"
    HPO = "HPO"
    OMIM = "OMIM"
    UNIPROT_PROTEIN_NAME = "UNIPROT_PROTEIN_NAME"
    UNIPROT_DOMAIN = "UNIPROT_DOMAIN"
    UNIPROT = "UNIPROT"
    MITOMAP = "MITOMAP"

    def is_a(self, a_dict):
        try:
            return a_dict[__TYPE__] == self.value
        except KeyError:
            return False

    def is_not(self, a_dict):
        return not self.is_a(a_dict)

    @classmethod
    def is_one(cls, doctypes, a_dict):
        """
        :param doctypes: iterable (preferably a set) of DocType values
        :param a_dict: object to check the type of
        :return:
        """

        try:
            return a_dict[__TYPE__] in doctypes
        except KeyError:
            return False

    @classmethod
    def get_doc_type(cls, a_dict):
        return a_dict.get(__CHILD__, a_dict.get(__TYPE__))

    @classmethod
    def get_default_columns(cls, a_dict):
        doc_type = cls.get_doc_type(a_dict)
        columns = DEFAULT_COLUMNS.get(doc_type)
        columns = columns or sorted(a_dict.keys())
        return [c for c in columns if not c.startswith("__")]

    @classmethod
    def get_header_names(cls, columns):
        return [to_header(c) for c in columns]

    @classmethod
    def get_sheet_name(cls, a_dict):
        doc_type = cls.get_doc_type(a_dict)
        return to_sheet_name(doc_type)

    @classmethod
    def create(cls, name: str):
        try:
            return DocType(name.upper() if name else None)
        except ValueError:
            pass


CALL_VARIANT_SET = {DocType.CALL.value, DocType.VARIANT.value}
CALL_VARIANT_KEYS = {"chr", "start", "ref", "alt"}


def is_call_or_variant(doc):
    return DocType.is_one(CALL_VARIANT_SET, doc) or not (
        CALL_VARIANT_KEYS - doc.keys()
    )


def is_not_skipped_call(doc):
    return DocType.SKIPPED_CALL.is_not(doc)


DEFAULT_COLUMNS = {
    DocType.GENE.value: ["chromosome", "start", "end", "name"],
    DocType.CALL.value: [
        "sample_name",
        "chr",
        "start",
        "ref",
        "alt",
        "ref_depth",
        "alt_depth",
        "vaf",
        "quality",
    ],
    DocType.ANNOTATED_CALL.value: [
        "sample_name",
        "annotations.hgvs_g",
        "annotations.canonical_c_dot",
        "annotations.canonical_alteration",
        "annotations.canonical_mutation_type",
        "annotations.canonical_gene",
        "annotations.GNOMAD__AF__mfloat",
        "annotations.ExAC__AF__float",
        "annotations.clinvar__CLNSIG__string",
        "ref_depth",
        "alt_depth",
        "vaf",
        "quality",
    ],
    DocType.ALTERATION.value: [
        "name",
        "mutation_type",
        "genes",
        "coordinates",
        "file_path",
    ],
}

COLUMN_TO_HEADER = {
    "annotations.hgvs_g": "hgvs (g.)",
    "annotations.canonical_c_dot": "hgvs (c.)",
    "annotations.canonical_alteration": "alteration",
    "annotations.canonical_mutation_type": "mutation type",
    "annotations.canonical_gene": "gene",
    "annotations.GNOMAD__AF__mfloat": "GNOMAD pop. freq.",
    "annotations.ExAC__AF__float": "ExAC pop. freq.",
}

TYPE_TO_SHEET = {}


def to_header(column):
    return COLUMN_TO_HEADER.setdefault(
        column, column.split(".")[-1].replace("_", " ")
    )


def to_sheet_name(doc_type):
    return TYPE_TO_SHEET.setdefault(
        doc_type, doc_type.replace("_", " ").title() + "s"
    )



from cytoolz.curried import map, curry
from .smart_lazy_file import SmartLazyFile

from genomoncology.pipeline.converters import (
    from_json_str,
    clean_str,
    non_null,
)


class Source(object):
    @staticmethod
    def is_source_class(obj):
        is_source = isinstance(obj, type) and issubclass(obj, Source)
        is_source = is_source or (
            isinstance(obj, curry)
            and isinstance(obj.func, type)
            and issubclass(obj.func, Source)
        )
        return is_source


class LazyFileSource(SmartLazyFile, Source):
    def __init__(self, filename):
        super().__init__(filename, mode="r")


class JsonlFileSource(LazyFileSource):
    def __iter__(self):
        iterator = super(JsonlFileSource, self).__iter__()
        return map(from_json_str, iterator)


class TextFileSource(LazyFileSource):
    def __iter__(self):
        iterator = super(TextFileSource, self).__iter__()
        return filter(non_null, map(clean_str, iterator))


class NullSource(LazyFileSource):
    """
    Returns back a single None to allow pipeline process to work
    for cases where records aren't being loaded like refresh_annotations.
    """

    def __iter__(self):
        return iter([None])



from cytoolz.curried import curry, assoc

from genomoncology.parse import DocType, __TYPE__, __CHILD__
from ..base import LazyFileSource

import sqlite3
import re
from tempfile import NamedTemporaryFile

from gosdk import logger


@curry
class MitomapFileSource(LazyFileSource):
    def __init__(self, dump_file_name, keep_db=False, **meta):
        self.meta = meta

        if __TYPE__ not in meta:
            self.meta = assoc(
                self.meta, __TYPE__, DocType.MITOMAP_RECORD.value
            )

        self.dump_file = open(dump_file_name, "r")
        self.table_names = [
            "genbank",
            "gbcontrol",
            "genbank_count",
            "genbank_haplogroup",
            "mitotip",
            "gbcontrol",
        ]
        self.table_indexed_fields = {
            "genbank": ["tpos", "qnt"],
            "gbcontrol": ["tpos", "qnt"],
            "genbank_count": ["haplogroup", "alt", "pos"],
            "genbank_haplogroup": ["haplogroup"],
            "mitotip": ["pos", "ref", "alt"],
        }
        self.lines = self.dump_file.readlines()
        self.table_line_boundaries = {}
        self.table_fields = {}
        # map of pos-ref-alt to set of haplogroups
        self.pra_map = {}
        # maps to gather counts during genbank table load
        self.var_haplo_cnt_map = {}
        self.haplo_cnt_map = {}
        self.fl_count_map = {}
        # set to get total fl count
        self.genbank_id_set = set()
        self.db = sqlite3.connect(":memory:")
        self.temp_file = NamedTemporaryFile(delete=(not keep_db))
        if keep_db:
            logger.get_logger().debug(
                "Mitomap database can be found at: {0}".format(
                    self.temp_file.name
                )
            )
        self.db = sqlite3.connect(self.temp_file.name)
        self.cursor = self.db.cursor()
        self.create_tables_in_db()

    # METHODS TO GET INFO ON TABLE START/END LINES
    def line_num_end_of_input(self, line_number, lines):
        for the_line in lines:
            if the_line.startswith("\\."):
                return line_number
            line_number += 1

    def get_table_line_boundaries(self, lines, starting_num, table_name):
        end_of_input = self.line_num_end_of_input(starting_num, lines)
        self.table_line_boundaries[table_name] = (starting_num, end_of_input)
        return end_of_input

    # METHODS TO CREATE SQL STATEMENTS
    def format_fields_with_data_types(self, fields):
        structured_fields = []
        for i in range(len(fields)):
            structured_fields.append(fields[i].strip() + " TEXT")
        return ",".join(structured_fields)

    def create_question_mark_string(self, length_of_fields):
        question_marks = ["?"] * length_of_fields
        return "{}".format(",".join(question_marks))

    def create_index_for_table(self, table_name, indexed_fields):
        return "CREATE INDEX index_{0} ON {0} ({1});".format(
            table_name, ",".join(indexed_fields)
        )

    def create_table_statement(self, table_name, fields):
        structured_fields = self.format_fields_with_data_types(fields)
        return "CREATE TABLE {0}({1});".format(table_name, structured_fields)

    def insert_into_table_statement(self, table_name, fields):
        return "INSERT INTO {0}({1}) VALUES({2});".format(
            table_name,
            ",".join(fields),
            self.create_question_mark_string(len(fields)),
        )

    def select_data_statement(self, table_name, selected_fields):
        return "SELECT {0} FROM {1};".format(
            ",".join(selected_fields), table_name
        )

    def select_cr_count(self, tpos, qnt):
        return (
            "SELECT count(*) FROM gbcontrol WHERE tpos='{0}' AND "
            "qnt='{1}';".format(tpos, qnt)
        )

    def select_total_count(self, table_name):
        return "SELECT count(DISTINCT genbank_id) FROM {0}".format(table_name)

    def select_variant_haplogroup_count(self, haplogroup, qnt, tpos):
        return (
            "SELECT cnt FROM genbank_count WHERE haplogroup='{0}' "
            "and alt='{1}' and pos='{2}';".format(haplogroup, qnt, tpos)
        )

    def select_haplogroup_count(self, haplogroup):
        return (
            "SELECT count(genbank_id) FROM genbank_haplogroup WHERE "
            "haplogroup='{0}';".format(haplogroup)
        )

    def select_mitotip(self, tpos, tnt, qnt):
        return (
            "SELECT score FROM mitotip WHERE pos='{0}' "
            "AND ref='{1}' AND alt='{2}'".format(tpos, tnt, qnt)
        )

    # METHODS TO EXECUTE SQL STATEMENTS
    def create_table(self, table_name, fields):
        self.cursor.execute(self.create_table_statement(table_name, fields))
        self.db.commit()

    def insert_into_table(
        self, table_name, fields, all_values, indexed_fields
    ):
        self.cursor.executemany(
            self.insert_into_table_statement(table_name, fields), all_values
        )
        if indexed_fields is not None:
            self.cursor.execute(
                self.create_index_for_table(table_name, indexed_fields)
            )
        self.db.commit()

    def get_table_data(self, table_name, selected_fields):
        self.cursor.execute(
            self.select_data_statement(table_name, selected_fields)
        )
        return self.cursor.fetchall()

    def get_fl_count(self, tpos, qnt):
        return self.fl_count_map[(tpos, qnt)]

    def get_cr_count(self, tpos, qnt):
        self.cursor.execute(self.select_cr_count(tpos, qnt))
        return self.cursor.fetchone()[0]

    def get_total_fl_count(self):
        return len(self.genbank_id_set)

    def get_total_count(self, table_name):
        self.cursor.execute(self.select_total_count(table_name))
        return self.cursor.fetchone()[0]

    def get_variant_haplogroup_count(self, haplogroup, qnt, tpos):
        self.cursor.execute(
            self.select_variant_haplogroup_count(haplogroup, qnt, tpos)
        )
        result = self.cursor.fetchone()
        if result is not None:
            return result[0]
        return 0  # return 0 if not present

    def get_haplogroup_count(self, haplogroup):
        self.cursor.execute(self.select_haplogroup_count(haplogroup))
        return self.cursor.fetchone()[0]

    def get_mitotip(self, tpos, tnt, qnt):
        self.cursor.execute(self.select_mitotip(tpos, tnt, qnt))
        result = self.cursor.fetchone()
        if result is not None:
            return result[0]
        return None  # return None if no score

    # METHODS TO CREATE AND POPULATE THE TABLES
    def parse_and_set_genbank_data(self, the_lines):
        data = []
        for line in the_lines:
            line = line.replace("\n", "")
            record = line.split("\t")
            # get the pos, ref, and alt to see if duplicate hgvs
            gb_fields = self.table_fields["genbank"]
            pos = record[gb_fields.index("tpos")]
            ref = record[gb_fields.index("tnt")]
            alt = record[gb_fields.index("qnt")]
            # get haplogroup
            haplogroup = record[gb_fields.index("haplogroup")]
            pra_key = (pos, ref, alt)
            if pra_key not in self.pra_map.keys():
                self.pra_map[pra_key] = {"haplogroup_set": set([haplogroup])}
                # add the first occurrence to the genbank table
                data.append(tuple(record))
            else:
                self.pra_map[pra_key]["haplogroup_set"].add(haplogroup)

            # calculate fl_count
            fl_key = (pos, alt)
            if fl_key not in self.fl_count_map.keys():
                self.fl_count_map[fl_key] = 1
            else:
                self.fl_count_map[fl_key] = self.fl_count_map[fl_key] + 1

            # add genbank id to set to use to calculate total fl count
            genbank_id = record[gb_fields.index("genbank_id")]
            self.genbank_id_set.add(genbank_id)

        return data

    def set_table_data(self, the_lines):
        data = []
        for line in the_lines:
            line = line.replace("\n", "")
            record = line.split("\t")
            data.append(tuple(record))
        return data

    def create_and_populate_table(
        self, table_name, the_lines, fields, indexed_fields
    ):
        if table_name == "genbank":
            data = self.parse_and_set_genbank_data(the_lines)
        else:
            data = self.set_table_data(the_lines)
        self.create_table(table_name, fields)
        self.insert_into_table(table_name, fields, data, indexed_fields)

    def create_tables_in_db(self):
        line_num = 0
        while line_num < len(self.lines):
            line = self.lines[line_num]
            if line.startswith("COPY") and line.split()[1] in self.table_names:
                table_name = line.split()[1]
                field_names = re.search(r"\((.*?)\)", line).group(1)
                self.table_fields[table_name] = list(
                    map(str.strip, field_names.split(","))
                )
                # get the line boundaries
                table_end_line = self.get_table_line_boundaries(
                    self.lines[line_num + 1 :], line_num + 1, table_name
                )
                # create and populate the table
                self.create_and_populate_table(
                    table_name,
                    self.lines[line_num + 1 : table_end_line],
                    self.table_fields[table_name],
                    self.table_indexed_fields[table_name],
                )
                line_num = table_end_line + 1
            line_num += 1

    def get_insertion(self, tnt, qnt):
        return qnt.replace(tnt, "", 1)

    def get_hgvs_g(self, tpos, tnt, qnt, ntchange):
        if ntchange == "insertion":
            insertion = self.get_insertion(tnt, qnt)
            return "NC_012920.1:g.{0}_{1}ins{2}".format(
                tpos, int(tpos) + 1, insertion
            )
        if ntchange == "deletion":
            return "NC_012920.1:g.{0}del{1}".format(tpos, tnt)
        return "NC_012920.1:g.{0}{1}>{2}".format(tpos, tnt, qnt)

    def get_freq(self, numer, denom):
        return 0 if denom == 0 else numer / denom

    def __iter__(self):  # pragma: no mccabe
        yield self.create_header()

        selected_fields = [
            "genbank_id",
            "tpos",
            "tnt",
            "qnt",
            "ntchange",
            "mmutid",
            "disease",
            "cal_aachange",
            "conservation",
            "haplogroup",
            "calc_locus",
        ]
        records = self.get_table_data("genbank", selected_fields)
        fl_total_count = self.get_total_fl_count()
        cr_total_count = self.get_total_count("gbcontrol")

        for rec in records:
            # get fields for each record
            tpos = rec[selected_fields.index("tpos")]
            tnt = rec[selected_fields.index("tnt")]
            qnt = rec[selected_fields.index("qnt")]
            ntchange = rec[selected_fields.index("ntchange")]
            hgvs_g = self.get_hgvs_g(tpos, tnt, qnt, ntchange)
            disease = rec[selected_fields.index("disease")]
            aa_change = rec[selected_fields.index("cal_aachange")]
            conservation = rec[selected_fields.index("conservation")]
            loci = rec[selected_fields.index("calc_locus")]
            # get haplogroups for this pos-ref-alt
            pra_key = (tpos, tnt, qnt)
            haplo_set = self.pra_map[pra_key]["haplogroup_set"]
            haplogroups = list(haplo_set)

            # get calculated fields
            fl_count = self.get_fl_count(tpos, qnt)
            cr_count = self.get_cr_count(tpos, qnt)
            fl_frequency = self.get_freq(fl_count, fl_total_count)
            cr_frequency = self.get_freq(cr_count, cr_total_count)
            mitoTIP = self.get_mitotip(tpos, tnt, qnt)

            var_haplo_counts = []
            haplo_counts = []
            for haplo in haplogroups:
                var_haplo_counts.append(
                    self.var_haplo_cnt_map.setdefault(
                        (haplo, qnt, tpos),
                        self.get_variant_haplogroup_count(haplo, qnt, tpos),
                    )
                )
                haplo_counts.append(
                    self.haplo_cnt_map.setdefault(
                        haplo, self.get_haplogroup_count(haplo)
                    )
                )

            yield {
                "hgvs_g": hgvs_g,
                "disease": disease,
                "aa_change": aa_change,
                "conservation": conservation,
                "haplogroups": haplogroups,
                "loci": loci,
                "fl_count": fl_count,
                "cr_count": cr_count,
                "fl_frequency": fl_frequency,
                "cr_frequency": cr_frequency,
                "variant_haplogroup_counts": var_haplo_counts,
                "haplogroup_counts": haplo_counts,
                "mitoTIP": mitoTIP,
                "__type__": DocType.MITOMAP_RECORD.value,
            }

        self.cursor.close()
        self.db.close()

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "meta": self.meta,
            "fields": [
                "hgvs_g",
                "disease",
                "aa_change",
                "conservation",
                "haplogroups",
                "loci",
                "fl_count",
                "cr_count",
                "fl_frequency",
                "cr_frequency",
                "variant_haplogroup_counts",
                "haplogroup_counts",
                "mitoTIP",
            ],
        }



from datetime import date

FIELD_TYPES_BY_CLASS = {int: "int", float: "float", date: "date"}


def get_field_type(v):
    prefix = ""

    if isinstance(v, list):
        prefix = "m"
        v = v[0] if len(v) > 0 else None

    field_type = FIELD_TYPES_BY_CLASS.get(v.__class__, "string")
    return f"{prefix}{field_type}"



from .base import Source, JsonlFileSource, TextFileSource, NullSource
from .bed import BedFileSource
from .delimited import DelimitedFileSource
from .paths import CollectFilePathsSource
from .excel import ExcelSource
from .aggregated import AggregatedFileSource
from .maf import MAFSource
from . import one_off_sources

__all__ = (
    "Source",
    "BedFileSource",
    "JsonlFileSource",
    "TextFileSource",
    "NullSource",
    "DelimitedFileSource",
    "CollectFilePathsSource",
    "ExcelSource",
    "AggregatedFileSource",
    "MAFSource",
)


def get_one_off_source(source_name):
    return one_off_sources.source_name_map.get(source_name)



from functools import partial
from typing import Iterable

from cytoolz.curried import curry, assoc

from gosdk import logger
from genomoncology.parse import (
    is_call_or_variant,
    __TYPE__,
    DocType,
    __CHILD__,
)
from genomoncology.parse.ensures import ensure_collection
from genomoncology.pipeline.converters import obj_to_dict
from gosdk.models import AnnotationList

from genomoncology.cli.const import GRCh37


@curry
async def async_annotate(
    data, sdk=None, fields=None, delete_if_exists=False, build=GRCh37
):
    csra_batch = convert_to_csra([{**d, "build": build} for d in data])

    annotation_bundle_version = await get_annotation_bundle_version_async(sdk)

    annotations_list = await sdk.call_with_retry(
        sdk.annotations.get_annotations,
        batch=csra_batch,
        delete_if_exists=delete_if_exists,
        build=build,
    )

    annotated_calls = annotate_calls(
        annotations_list, data, fields, annotation_bundle_version
    )

    logger.get_logger().debug(
        "async_annotate",
        call_count=len(data),
        first_call=data[:1],
        csra_batch=csra_batch[:4],
        annotated_count=len(annotated_calls),
        first_annotated=annotated_calls[:1],
    )

    return annotated_calls


@curry
def sync_annotate(
    data, sdk=None, fields=None, delete_if_exists=False, build=GRCh37
):
    csra_batch = convert_to_csra([{**d, "build": build} for d in data])

    annotation_bundle_version = get_annotation_bundle_version(sdk)

    annotations_list = (
        sdk.call_with_retry(
            sdk.annotations.get_annotations,
            batch=csra_batch,
            delete_if_exists=delete_if_exists,
            build=build,
        )
        if csra_batch
        else None
    )

    annotated_calls = annotate_calls(
        annotations_list, data, fields, annotation_bundle_version
    )

    logger.get_logger().debug(
        "sync_annotate",
        call_count=len(data),
        call_first=data[:1],
        csra_batch=csra_batch[:4] if csra_batch else None,
        annotated_count=len(annotated_calls),
        annotated_first=annotated_calls[:1],
    )

    return annotated_calls


@curry
async def async_load_annotations(
    data, sdk=None, data_set=None, data_set_version=None, build=GRCh37
):
    await sdk.call_with_retry(
        sdk.annotations.load_annotations,
        data=override_hgvs([{**d, "build": build} for d in data]),
        data_set=data_set,
        data_set_version=data_set_version,
    )

    logger.get_logger().debug(
        "async_load_annotations", call_count=len(data), first_record=data[:1]
    )

    # response of warehouse is blank, returning data for downstream processing
    return data


@curry
def sync_refresh(data, sdk=None, recalculate_hgvs=False, batch_size=50):
    chromosome, start, end = format_chromosome_start_end(data)

    cursor_mark = "*"
    has_more = True

    while has_more:
        future = sdk.annotations.refresh_annotations(
            cursor_mark=cursor_mark,
            rows=batch_size,
            recalculate_hgvs=recalculate_hgvs,
            chromosome=chromosome,
            start=start,
            end=end,
        )

        result = future.result()
        has_more = result.more
        cursor_mark = result.next_cursor_mark

        # echo message
        msg = [result.rows, has_more, result.last_hgvs, cursor_mark]
        print("  /  ".join(map(str, msg)))


# private methods
def override_hgvs(data):
    data = ensure_collection(data)
    return list(map(up_build, data))


def convert_to_csra(data):
    # create batch of CSRA strings for calling HGVS API
    data = ensure_collection(data)
    calls_only = filter(is_call_or_variant, data)
    csra_batch = list(map(to_csra, calls_only))
    return csra_batch


def annotate_calls(annotations_list, data, fields, annotation_bundle_version):
    # todo: determine if it makes sense to not return all annotations....
    fields = None  # resolve_fields(fields)
    func = partial(
        add_annotation, annotations_list, fields, annotation_bundle_version
    )
    annotated_calls = list(map(func, data))
    return annotated_calls


def add_annotation(
    annotations_list: AnnotationList,
    fields,
    expected_bundle_version,
    call: dict,
):

    if annotations_list and is_call_or_variant(call):
        csra = to_csra(call)
        annotations = obj_to_dict(
            fields, annotations_list.get_annotation(csra)
        ) or {}

        actual_bundle_version = annotations.get(
            "annotation_bundle_version",
            expected_bundle_version
        ) or "unspecified_version"

        # if the annotation_bundle does not match, raise an exception
        if expected_bundle_version != actual_bundle_version:
            msg = (
                "Annotations Process Terminated due to "
                "annotations_bundle_version "
                "changing during a run. \n"
                "Expected annotation bundle version: {expected} \n"
                "Actual annotation bundle version: {actual}".format(
                    expected=expected_bundle_version,
                    actual=actual_bundle_version,
                )
            )
            raise AnnBundleVersionAssertionError(msg)

        call["annotations"] = annotations
        call[__TYPE__] = f"ANNOTATED_{call.get(__TYPE__, 'CALL')}"

    elif DocType.HEADER.is_a(call):
        call[__CHILD__] = f"ANNOTATED_{call.get(__CHILD__, 'CALL')}"

    return call


def up_build(call: dict):
    hgvs_g = call.get("hgvs_g")
    build = call.get("build")
    has_both = hgvs_g is not None and build is not None

    if has_both and not hgvs_g.endswith(f"|{build}"):
        hgvs_parts = hgvs_g.split("|")
        if len(hgvs_parts) >= 4:
            hgvs_parts = hgvs_parts[:4] + [build]
            call = assoc(call, "hgvs_g", "|".join(hgvs_parts))

    return call


def to_csra(call: dict):
    return "|".join(
        (
            f"chr{call.get('chr')}",
            str(call.get("start")),
            call.get("ref") or "-",
            call.get("alt") or "-",
            call.get("build") or GRCh37,
        )
    )


def resolve_fields(fields: Iterable[str] = None) -> Iterable[str]:
    fields = ensure_collection(fields)
    if fields == ["*"]:
        fields = None
    else:
        fields = ANNOTATIONS + [f for f in fields if f not in ANNOTATIONS_SET]
    return fields


def format_chromosome_start_end(data):
    if data is None:
        return None, None, None
    data = [None if not str(x).strip() else x for x in data]
    return data[0], data[1], data[2]


async def get_annotation_bundle_version_async(sdk):
    annotation_bundle_version = await sdk.call_with_retry(
        sdk.info.get_annotation_bundle_version
    )
    return annotation_bundle_version.annotation_bundle_version


def get_annotation_bundle_version(sdk):
    annotation_bundle_version = sdk.call_with_retry(
        sdk.info.get_annotation_bundle_version
    )
    return annotation_bundle_version.annotation_bundle_version


class AnnBundleVersionAssertionError(AssertionError):
    pass


ANNOTATIONS = [
    "GNOMAD__AF__mfloat",
    "ExAC__AF__float",
    "canonical_alteration",
    "canonical_c_dot",
    "canonical_gene",
    "canonical_mutation_type",
    "canonical_p_dot",
    "clinvar__CLNSIG__string",
    "display",
    "mutation_type",
    "representations",
    "hgvs_g",
    "alteration",
]

ANNOTATIONS_SET = set(ANNOTATIONS)



import xml.etree.cElementTree as etree

from cytoolz.curried import curry, assoc

from genomoncology.parse import DocType, __TYPE__, __CHILD__
from ..base import LazyFileSource


@curry
class UniprotFileSource(LazyFileSource):
    def __init__(self, filename, **meta):
        self.filename = filename
        self.meta = meta

        if __TYPE__ not in meta:
            self.meta = assoc(
                self.meta, __TYPE__, DocType.UNIPROT_RECORD.value
            )

    def __iter__(self):  # pragma: no mccabe
        yield self.create_header()

        context = etree.iterparse(self.filename, events=["start", "end"])
        # accession
        accessions = []
        # protein name
        in_protein_rec_name = False
        protein = None
        # gene name
        in_gene_name = False
        gene = None
        # domain
        in_domain = False
        domains = []
        domain_starts = []
        domain_ends = []
        # protein length
        protein_length = None
        # np_ids
        np_ids = []
        for event, elem in context:
            # accession
            if event == "start" and elem.tag == "accession":
                accessions.append(elem.text)
                continue
            # protein name
            if event == "start" and elem.tag == "recommendedName":
                in_protein_rec_name = True
                continue
            if (
                in_protein_rec_name
                and event == "end"
                and elem.tag == "fullName"
            ):
                protein = elem.text
                in_protein_rec_name = False
                continue
            # gene name
            if event == "start" and elem.tag == "gene":
                in_gene_name = True
                continue
            if (
                in_gene_name
                and event == "end"
                and elem.tag == "name"
                and "type" in elem.attrib
                and elem.attrib["type"] == "primary"
            ):
                gene = elem.text
                in_gene_name = False
                continue
            # domains
            if (
                event == "start"
                and elem.tag == "feature"
                and "type" in elem.attrib
                and elem.attrib["type"] == "domain"
            ):
                domains.append(elem.attrib["description"])
                in_domain = True
                continue
            if in_domain and event == "start" and elem.tag == "begin":
                domain_starts.append(elem.attrib["position"])
                continue
            if in_domain and event == "start" and elem.tag == "end":
                domain_ends.append(elem.attrib["position"])
                continue
            if (
                event == "end"
                and elem.tag == "feature"
                and "type" in elem.attrib
                and elem.attrib["type"] == "domain"
            ):
                in_domain = False
                continue
            # protein length
            if (
                event == "start"
                and elem.tag == "sequence"
                and "length" in elem.attrib
            ):
                protein_length = elem.attrib["length"]
                continue
            # np_ids
            if (
                event == "start"
                and elem.tag == "dbReference"
                and elem.attrib["type"] == "RefSeq"
            ):
                np_ids.append(elem.attrib["id"])
                continue
            # end of entry record
            if event == "end" and elem.tag == "entry":
                return_d = {
                    "accessions": accessions,
                    "gene": gene,
                    "protein_full_name": protein,
                    "domains": domains,
                    "domain_starts": domain_starts,
                    "domain_ends": domain_ends,
                    "protein_length": protein_length,
                    "np_ids": np_ids,
                    "__type__": DocType.UNIPROT_RECORD,
                }
                accessions = []
                gene = None
                protein = None
                domains = []
                domain_starts = []
                domain_ends = []
                protein_length = None
                np_ids = []
                elem.clear()
                yield return_d

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "meta": self.meta,
            "fields": [
                "gene",
                "protein_full_name",
                "accessions",
                "domains",
                "domain_starts",
                "domain_ends",
                "protein_length",
                "np_ids",
            ],
        }



import asyncio
from multiprocessing import Process, Queue

from cytoolz.curried import curry
from cytoolz.curried import dissoc

from genomoncology.parse import DocType
from genomoncology.parse import ensures
from . import Sink

EOL = "EOL"


async def make_call(sdk, sdk_func, in_queue):
    unit = in_queue.get()

    while unit != EOL:
        await sdk.call_with_retry(sdk_func, data=unit)
        unit = in_queue.get()


def background_runner(state, data_type, queue):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    sdk = state.create_sdk(async_enabled=True, loop=loop)
    sdk_module = getattr(sdk, f"warehouse_{data_type}")
    sdk_func = getattr(sdk_module, f"create_warehouse_{data_type}")
    loop.run_until_complete(asyncio.gather(make_call(sdk, sdk_func, queue)))


class LoadWarehouseSink(Sink):

    DATA_TYPE = None

    def __init__(self, _, state, num_workers=10):
        assert self.DATA_TYPE, "Do not instantiate LoadWarehouseSink directly."

        self.queue = Queue(maxsize=1000)
        self.workers = []

        self.num_workers = num_workers
        self.build = state.build

        for _ in range(num_workers):
            worker = Process(
                target=background_runner,
                args=(state, self.DATA_TYPE, self.queue),
            )
            worker.start()
            self.workers.append(worker)

    def close(self):
        for _ in range(self.num_workers):
            self.queue.put(EOL)

        for worker in self.workers:
            worker.join()

    def write(self, data):
        data = ensures.ensure_collection(data)
        data = filter(DocType.HEADER.is_not, data)
        data = [
            dissoc(d, "__type__", "__record__", "annotations") for d in data
        ]
        data = [{**d, "build": self.build} for d in data]

        self.queue.put(data)

        return data


@curry
class LoadWarehouseVariantsSink(LoadWarehouseSink):

    DATA_TYPE = "variants"


@curry
class LoadWarehouseFeaturesSink(LoadWarehouseSink):

    DATA_TYPE = "features"



from cytoolz.curried import assoc, compose, remove, pipe, merge_with, merge
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    "gene": "gene",
    "gene_mim__string": "Mim Number",
    "phenotype__mstring": "phenotype",
    "phenotype_mim__mstring": "phenotype_mim",
    "inheritance__mstring": "inheritance",
}


def split_filter(value, delimiter, pred_func):
    return filter(pred_func, value.split(delimiter))


def split_remove(value, delimiter, pred_func):
    return remove(pred_func, value.split(delimiter))


def parse_omim_inheritance(omim_pheno_d):
    omim_inheritance_l = [
        "Multifactorial",
        "autosomal dominant",
        "X-linked dominant",
        "X-linked",
        "X-linked recessive",
        "Y-linked",
        "autosomal recessive",
        "Autosomal dominant",
        "Autosomal recessive",
    ]
    omim_pheno_d = assoc(
        omim_pheno_d,
        "inheritance",
        ";".join(
            split_filter(
                omim_pheno_d["phenotype"],
                ", ",
                lambda x: x in omim_inheritance_l,
            )
        ),
    )
    omim_pheno_d = assoc(
        omim_pheno_d,
        "phenotype",
        ", ".join(
            split_remove(
                omim_pheno_d["phenotype"],
                ", ",
                lambda x: x in omim_inheritance_l,
            )
        ),
    )
    return omim_pheno_d


def parse_omim_position(omim_pheno_d):
    omim_position_l = ["(1)", "(2)", "(3)", "(4)"]
    omim_pheno_d = assoc(
        omim_pheno_d,
        "phenotype",
        " ".join(
            split_remove(
                omim_pheno_d["phenotype"], " ", lambda x: x in omim_position_l
            )
        ),
    )
    return omim_pheno_d


def parse_omim_pheno_id(omim_pheno_d):
    if omim_pheno_d["phenotype"][-6:].isdigit():
        omim_pheno_d = assoc(
            omim_pheno_d, "phenotype_mim", omim_pheno_d["phenotype"][-6:]
        )
        omim_pheno_d = assoc(
            omim_pheno_d, "phenotype", omim_pheno_d["phenotype"][:-8]
        )
    else:
        omim_pheno_d = assoc(omim_pheno_d, "phenotype_mim", "")
    return omim_pheno_d


def phenotype_pipeline(x):
    return pipe(
        {"phenotype": x},
        parse_omim_inheritance,
        parse_omim_position,
        parse_omim_pheno_id,
    )


def parse_omim_phenotypes(x):
    if "Phenotypes" in x:
        return merge_with(
            list, map(phenotype_pipeline, x["Phenotypes"].split("; "))
        )
    else:
        return x


def parse_omim_gene(x):
    return x["Gene Symbols"].split(", ")[0]


register(
    input_type=DocType.TSV,
    output_type=DocType.OMIM,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.OMIM.value),
        name_mapping(NAME_MAPPING),
        lambda x: merge(x, parse_omim_phenotypes(x)),
        lambda x: assoc(x, "gene", parse_omim_gene(x)),
        # name_mapping(NAME_MAPPING_PRE),
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.OMIM,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.OMIM.value),
        # lambda x: merge(x, parse_omim_phenotypes(x)),
    ),
    is_header=True,
)



from click.types import ParamType
import os

DATASETS_DIR = os.path.join(os.path.dirname(__file__), "../../datasets/")


class DatasetPath(ParamType):
    """
    A dataset can be one of three types of values:
        - local, relative file path (e.g. ./subdir/my_file.txt)
        - local, absolute file path (e.g. /opt/subdir/my_file.txt)
        - gocli dataset src/datasets (e.g. @snv_cancer.bed)
    """

    def convert(self, value, param, ctx):
        if value.startswith("@"):
            value = os.path.join(DATASETS_DIR, value[1:])
            print(f"HERE! {value}")

        return value

        # # if a full path, leave it, otherwise get from src/datasets
        # if not os.path.exists(file_path):
        #     file_path = os.path.join(DATASETS_DIR, file_path)
        #
        # assert os.path.exists(file_path), f"{file_path} not found."

        #
        # try:
        #     st = os.stat(rv)
        # except OSError:
        #     if not self.exists:
        #         return self.coerce_path_result(rv)
        #     self.fail('%s "%s" does not exist.' % (
        #         self.path_type,
        #         filename_to_ui(value)
        #     ), param, ctx)



from .utils import int_or_none


class BaseInfoMapper(object):
    """
    Map INFO and FORMAT data to 3 key values: total, alt, and ref depth.
    """

    @classmethod
    def keys(cls):
        raise NotImplementedError

    def do_mapping(self, call_info, allele_position, alt_allele):
        raise NotImplementedError

    def calculate_vafs(self, total_depth, alt_depth, ref_depth):
        vaf, vaf_alt, ref_plus_alt = None, None, None

        alt_depth = int_or_none(alt_depth)
        ref_depth = int_or_none(ref_depth)
        total_depth = int_or_none(total_depth)

        if alt_depth is not None and ref_depth is not None:
            ref_plus_alt = alt_depth + ref_depth

        total_depth = total_depth if total_depth is not None else ref_plus_alt

        if total_depth is not None and alt_depth is not None:
            vaf = (float(alt_depth) / total_depth) if alt_depth else None
            vaf_alt = (
                (float(alt_depth) / ref_plus_alt) if ref_plus_alt else vaf
            )

        return vaf, vaf_alt

    def __call__(self, call_info, allele_position, alt_allele):
        total_depth, alt_depth, ref_depth = self.do_mapping(
            call_info, allele_position, alt_allele
        )

        vaf, vaf_alt = self.calculate_vafs(total_depth, alt_depth, ref_depth)

        return dict(
            total_depth=total_depth,
            alt_depth=alt_depth,
            ref_depth=ref_depth,
            vaf=vaf,
            vaf_alt=vaf_alt,
        )


all_mapping = []
lookup = {}


def register(cls, priority=0):
    """ register a given mapper class based it's priority. """
    cls_name = cls.__name__
    lookup[cls_name] = cls
    all_mapping.append((priority, cls_name, cls.keys()))
    all_mapping.sort()


def find(vcf_keys):
    vcf_keys = set(vcf_keys)
    matches = []
    for (priority, cls_name, keys) in all_mapping:
        if keys.issubset(vcf_keys):
            matches.append((priority, len(keys), cls_name))

    if matches:
        cls_name = sorted(matches, reverse=True)[0][-1]
        cls = lookup[cls_name]
        return cls()



from cytoolz.curried import curry
from glom import Coalesce

from genomoncology.kms.match import (
    calls_to_alterations,
    transform_match_response,
)
from genomoncology.parse import DocType, __TYPE__
from genomoncology.parse.ensures import ensure_collection


@curry
def match_contents(calls, diseases=None, include_content=True, sdk=None):
    alterations = calls_to_alterations(calls)
    diseases = ensure_collection(diseases or ["ANY"])

    response = sdk.call_with_retry(
        sdk.contents.match_contents_post,
        alterations=alterations,
        diseases=diseases,
        include_content=include_content,
    )

    return transform_match_response(response, DEFAULT_SPEC)


DEFAULT_SPEC = {
    __TYPE__: Coalesce(default=DocType.CONTENT.value),
    "title": "title",
    "body": "body",
    "url": lambda content: f"https://www.mycancergenome.org{content.url}",
    "type": "type",
    "trigger_alterations": "trigger_alterations",
    "trigger_genes": "trigger_genes",
    "trigger_diseases": "trigger_diseases",
}



from genomoncology.parse.ensures import flatten
from cytoolz.curried import curry
from glom import glom

from .doctypes import DocType


class TuxedoMapper(object):
    def __init__(self, bed_filter):
        self.bed_filter = bed_filter
        self.base = dict(__type__=DocType.ALTERATION.value)
        self.seen = set()

    def get_genes(self, chromosome, position):
        tree = self.bed_filter.get_tree(chromosome)
        genes = []
        if tree:
            genes = [interval.data for interval in tree.search(position)]
        return flatten(genes) or [None]

    def add_meta(self, fusion, name, value):
        fusion.setdefault("meta", {})[name] = value

    def construct_fusion(self, *genes):
        genes = list(filter(None, genes))

        if genes:
            fusion = self.base.copy()
            if len(genes) == 1 or genes[0] != genes[1]:
                fusion["mutation_hash"] = "{} Fusion".format("-".join(genes))
            else:
                genes = list(set(genes))
                assert len(genes) == 1, f"Expected inversion: {genes}"
                fusion["mutation_hash"] = "{} Inversion".format(
                    "-".join(genes)
                )

            self.add_meta(fusion, "genes", genes)
            return fusion

    def __call__(self, doc):  # pragma: no mccabe
        if DocType.HEADER.is_a(doc):
            self.base["file_path"] = doc.get("file_path", None)
            self.seen.clear()

        elif DocType.TUXEDO.is_a(doc):
            chr_1, chr_2 = doc.get("chr_1"), doc.get("chr_2")
            pos_1, pos_2 = doc.get("pos_1"), doc.get("pos_2")

            for gene_1 in self.get_genes(chr_1, pos_1):
                for gene_2 in self.get_genes(chr_2, pos_2):
                    fusion = self.construct_fusion(gene_1, gene_2)
                    if fusion and fusion["mutation_hash"] not in self.seen:
                        self.seen.add(fusion["mutation_hash"])

                        for (name, value) in doc.items():
                            if not name.startswith("__"):
                                self.add_meta(fusion, name, value)

                        yield fusion


DEFAULT_TUXEDO_FILE_COLUMNS = [
    "chromosomes",
    "pos_1",
    "pos_2",
    "orientation",
    "spanning",
    "mate_pairs",
    "mate_pairs_and_spanning_one",
    "contradictory_reads",
]

SPEC = {
    "__type__": "__type__",
    "chr_1": ("chromosomes", lambda x: x.split("-")[0].replace("chr", "")),
    "chr_2": ("chromosomes", lambda x: x.split("-")[1].replace("chr", "")),
    "pos_1": ("pos_1", int),
    "pos_2": ("pos_2", int),
    "orientation": "orientation",
    "spanning": ("spanning", int),
    "mate_pairs": ("mate_pairs", int),
    "mate_pairs_and_spanning_one": ("mate_pairs_and_spanning_one", int),
    "contradictory_reads": ("contradictory_reads", int),
}


@curry
def iterate_tuxedo_files(file_paths, columns=None):
    columns = columns or DEFAULT_TUXEDO_FILE_COLUMNS

    for file_path in file_paths:

        # todo: refactor: cannot depend on pipeline!
        from genomoncology.pipeline.sources import DelimitedFileSource

        source = DelimitedFileSource(
            file_path,
            delimiter="\t",
            columns=columns,
            __type__=DocType.TUXEDO.value,
        )

        for record in source:
            if DocType.HEADER.is_a(record):
                yield record
            else:
                transformed = glom(record, SPEC)
                yield transformed



from cytoolz.curried import filter, map
from cytoolz.functoolz import compose
from cytoolz.itertoolz import concat

from genomoncology.parse.ensures import ensure_collection
from genomoncology.pipeline.converters import non_null
from genomoncology.pipeline.transformers import get_in_field, transform


def calls_to_alterations(calls):
    calls = ensure_collection(calls)
    pipeline = compose(
        concat, filter(non_null), map(get_in_field("annotations.alteration"))
    )
    alterations = list(pipeline(calls))
    return alterations


def transform_match_response(response, default_spec):
    transformer = transform(default_spec)

    # handle either model or dict version of response
    # contents -> object  //  everything else -> dict
    results = getattr(response, "results", None)
    results = results or response.get("results")

    transformed = map(transformer, results)
    return transformed



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    "gene": "gene",
    # info
    "phenotype__mstring": "phenotype",
    "hpo__mstring": "hpo_id",
}

register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.HPO,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.HPO.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.HPO,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.HPO.value),
    ),
    is_header=True,
)



from genomoncology import kms
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    # hgvs
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    "build": "build",
    # info
    "AC__mint": "info.AC",
    "AF__mfloat": "info.AF",
    "AN__mint": "info.AN",
    "BaseQRankSum__mfloat": "info.BaseQRankSum",
    "ClippingRankSum__mfloat": "info.ClippingRankSum",
    "DB__mstring": "info.DB",
    "DP__mint": "info.DP",
    "FS__mfloat": "info.FS",
    "InbreedingCoeff__mfloat": "info.InbreedingCoeff",
    "MQ__mfloat": "info.MQ",
    "MQRankSum__mfloat": "info.MQRankSum",
    "QD__mfloat": "info.QD",
    "ReadPosRankSum__mfloat": "info.ReadPosRankSum",
    "SOR__mfloat": "info.SOR",
    "VQSR_culprit__mstring": "info.VQSR_culprit",
    "VQSR_NEGATIVE_TRAIN_SITE__mstring": "info.VQSR_NEGATIVE_TRAIN_SITE",
    "VQSR_POSITIVE_TRAIN_SITE__mstring": "info.VQSR_POSITIVE_TRAIN_SITE",
    "GQ_HIST_ALT__mstring": "info.GQ_HIST_ALT",
    "DP_HIST_ALT__mstring": "info.DP_HIST_ALT",
    "AB_HIST_ALT__mstring": "info.AB_HIST_ALT",
    "GQ_HIST_ALL__mstring": "info.GQ_HIST_ALL",
    "DP_HIST_ALL__mstring": "info.DP_HIST_ALL",
    "AB_HIST_ALL__mstring": "info.AB_HIST_ALL",
    "AC_AFR__mint": "info.AC_AFR",
    "AC_AMR__mint": "info.AC_AMR",
    "AC_ASJ__mint": "info.AC_ASJ",
    "AC_EAS__mint": "info.AC_EAS",
    "AC_FIN__mint": "info.AC_FIN",
    "AC_NFE__mint": "info.AC_NFE",
    "AC_OTH__mint": "info.AC_OTH",
    "AC_Male__mint": "info.AC_Male",
    "AC_Female__mint": "info.AC_Female",
    "AN_AFR__mint": "info.AN_AFR",
    "AN_AMR__mint": "info.AN_AMR",
    "AN_ASJ__mint": "info.AN_ASJ",
    "AN_EAS__mint": "info.AN_EAS",
    "AN_FIN__mint": "info.AN_FIN",
    "AN_NFE__mint": "info.AN_NFE",
    "AN_OTH__mint": "info.AN_OTH",
    "AN_Male__mint": "info.AN_Male",
    "AN_Female__mint": "info.AN_Female",
    "AF_AFR__float": "info.AF_AFR",
    "AF_AMR__mfloat": "info.AF_AMR",
    "AF_ASJ__mfloat": "info.AF_ASJ",
    "AF_EAS__mfloat": "info.AF_EAS",
    "AF_FIN__mfloat": "info.AF_FIN",
    "AF_NFE__mfloat": "info.AF_NFE",
    "AF_OTH__mfloat": "info.AF_OTH",
    "AF_Male__mfloat": "info.AF_Male",
    "AF_Female__mfloat": "info.AF_Female",
    "GC_AFR__mint": "info.GC_AFR",
    "GC_AMR__mint": "info.GC_AMR",
    "GC_ASJ__mint": "info.GC_ASJ",
    "GC_EAS__mint": "info.GC_EAS",
    "GC_FIN__mint": "info.GC_FIN",
    "GC_NFE__mint": "info.GC_NFE",
    "GC_OTH__mint": "info.GC_OTH",
    "GC_Male__mint": "info.GC_Male",
    "GC_Female__mint": "info.GC_Female",
    "AC_raw__mint": "info.AC_raw",
    "AN_raw__mint": "info.AN_raw",
    "AF_raw__mfloat": "info.AF_raw",
    "GC_raw__mint": "info.GC_raw",
    "GC__mint": "info.GC",
    "Hom_AFR__mint": "info.Hom_AFR",
    "Hom_AMR__mint": "info.Hom_AMR",
    "Hom_ASJ__mint": "info.Hom_ASJ",
    "Hom_EAS__mint": "info.Hom_EAS",
    "Hom_FIN__mint": "info.Hom_FIN",
    "Hom_NFE__mint": "info.Hom_NFE",
    "Hom_OTH__mint": "info.Hom_OTH",
    "Hom_Male__mint": "info.Hom_Male",
    "Hom_Female__mint": "info.Hom_Female",
    "Hom_raw__mint": "info.Hom_raw",
    "Hom__mint": "info.Hom",
    "STAR_AC__mint": "info.STAR_AC",
    "STAR_AC_raw__mint": "info.STAR_AC_raw",
    "STAR_Hom__mint": "info.STAR_Hom",
    "POPMAX__mstring": "info.POPMAX",
    "AC_POPMAX__mint": "info.AC_POPMAX",
    "AN_POPMAX__mint": "info.AN_POPMAX",
    "AF_POPMAX__mfloat": "info.AF_POPMAX",
    "DP_MEDIAN__mint": "info.DP_MEDIAN",
    "DREF_MEDIAN__mfloat": "info.DREF_MEDIAN",
    "GQ_MEDIAN__mfloat": "info.GQ_MEDIAN",
    "AB_MEDIAN__mfloat": "info.AB_MEDIAN",
    "AS_RF__mfloat": "info.AS_RF",
    "AS_FilterStatus__mstring": "info.AS_FilterStatus",
    "AS_RF_POSITIVE_TRAIN__mint": "info.AS_RF_POSITIVE_TRAIN",
    "AS_RF_NEGATIVE_TRAIN__mint": "info.AS_RF_NEGATIVE_TRAIN",
    "CSQ__mstring": "info.CSQ",
}

register(
    input_type=DocType.CALL,
    output_type=DocType.GNOMAD,
    transformer=compose(
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        lambda x: assoc(x, __TYPE__, DocType.GNOMAD.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.CALL,
    output_type=DocType.GNOMAD,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.GNOMAD.value)),
    is_header=True,
)



from cytoolz.curried import curry

from genomoncology.pipeline.sources.base import LazyFileSource


@curry
class BedFileSource(LazyFileSource):
    def __init__(self, _, bed_file, **meta):
        assert bed_file, "BED not provided"
        self.file = open(bed_file, "r")
        self.entries = self.file.readlines()
        self.meta = meta

    def __iter__(self):
        for line in self.entries:
            data = line.split("\t")
            if len(data) > 2:
                chromosome = data[0]
                start = data[1]
                end = data[2]
                yield (chromosome, start, end)

        self.file.close()



from cytoolz.curried import curry, assoc
from cytoolz import reduceby

from genomoncology.parse import DocType, __TYPE__, __CHILD__
from .base import LazyFileSource
from .delimited import do_split


@curry
class AggregatedFileSource(LazyFileSource):
    def __init__(self, filename, aggregate_key, delimiter="\t", **meta):
        self.delimiter = delimiter
        self.aggregate_key = aggregate_key
        self.meta = meta

        if __TYPE__ not in meta:
            self.meta = assoc(self.meta, __TYPE__, DocType.AGGREGATE.value)

        super().__init__(filename)

    def __iter__(self):
        # noinspection PyUnresolvedReferences
        iterator = super(AggregatedFileSource.func, self).__iter__()

        self.columns = next(iterator).strip().split(self.delimiter)

        yield self.create_header()

        aggregated_d = reduceby(
            self.get_key_value, self.get_aggregate_value, iterator, dict
        )

        for key, value in aggregated_d.items():
            value["gene"] = key
            value["__type__"] = DocType.AGGREGATE.value
            yield value

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "columns": self.columns,
            "meta": self.meta,
            "file_path": self.name,
        }

    def get_key_value(self, x):
        key = do_split(self.delimiter, x.replace("\n", ""))[
            self.columns.index(self.aggregate_key)
        ]
        return key

    def get_aggregate_value(self, acc, x):
        hold_d = acc
        value_l = do_split(self.delimiter, x.replace("\n", ""))
        for i in range(len(value_l)):
            value = value_l[i] if value_l[i] != "" else "None"
            if self.columns[i] in hold_d:
                hold_d[self.columns[i]] = hold_d[self.columns[i]] + [value]
            else:
                hold_d[self.columns[i]] = [value]
        return hold_d



from click.utils import LazyFile
from cytoolz.curried import curry, dissoc

from genomoncology.pipeline.converters import to_json_str


class Sink(object):
    def write(self, unit):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    @staticmethod
    def is_sink_class(obj):
        is_sink = isinstance(obj, type) and issubclass(obj, Sink)
        is_sink = is_sink or (
            isinstance(obj, curry)
            and isinstance(obj.func, type)
            and issubclass(obj.func, Sink)
        )
        return is_sink


class FileSink(Sink):
    def __init__(self, filename, insert_newlines=True):
        self.output = LazyFile(filename, mode="w")
        self.insert_newlines = insert_newlines

    def convert(self, unit):
        raise NotImplementedError

    def write(self, unit):
        if unit is not None:

            unit = self.convert(unit)

            if unit is not None:
                self.output.write(unit)

                if self.insert_newlines:
                    self.output.write("\n")

        return unit

    def close(self):
        if self.output.name != "-":
            self.output.close()


class JsonlFileSink(FileSink):
    def convert(self, unit):
        return to_json_str(unit)


class TextFileSink(FileSink):
    def convert(self, unit):
        return str(dissoc(unit, "__record__"))



# Shared state and options for all commands.
#
# Reference:
# https://github.com/pallets/click/issues/108#issuecomment-44691173

import uuid

import click
from gosdk import variables

from genomoncology import State
from . import const

pass_state = click.make_pass_decorator(State, ensure=True)


def debug_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.debug = value
        return value

    return click.option(
        "-d",
        "--debug",
        is_flag=True,
        expose_value=False,
        help="Run in debug mode. (debug level logging).",
        callback=callback,
    )(f)


def quiet_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.quiet = value
        return value

    return click.option(
        "-q",
        "--quiet",
        is_flag=True,
        expose_value=False,
        help="Run in quiet mode. (warning level logging).",
        callback=callback,
    )(f)


def pipeline_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.pipeline = value
        return value

    return click.option(
        "-p",
        "--pipeline",
        type=str,
        expose_value=False,
        help="Pipeline identifier.",
        callback=callback,
    )(f)


def build_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.build = value
        return value

    return click.option(
        "-b",
        "--build",
        type=click.Choice(const.BUILDS),
        expose_value=False,
        help="Reference assembly.",
        callback=callback,
        default=const.GRCh37,
    )(f)


def run_id_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.run_id = value or str(uuid.uuid4())
        return value

    return click.option(
        "-r",
        "--run_id",
        type=str,
        expose_value=False,
        help="Unique identifier of this run.",
        callback=callback,
    )(f)


def size_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.batch_size = value
        return value

    return click.option(
        "-s",
        "--size",
        type=int,
        default=50,
        expose_value=False,
        help="Size of batches when calling API.",
        callback=callback,
    )(f)


def glob_option(f, default=("*.vcf", "*.vcf.gz")):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.glob = value
        return value

    return click.option(
        "--glob",
        "-g",
        help="File pattern (e.g. *.vcf) of files to load",
        multiple=True,
        type=str,
        default=default,
        expose_value=False,
        callback=callback,
    )(f)


def include_tar_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.include_tar = value
        return value

    return click.option(
        "--include-tar",
        is_flag=True,
        expose_value=False,
        help="Search in tar (tgz, tar.gz, tar) files",
        callback=callback,
    )(f)


def hard_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        state.hard_failure = value
        return value

    return click.option(
        "-h",
        "--hard",
        is_flag=True,
        expose_value=False,
        help="Hard stop on failure.",
        callback=callback,
    )(f)


def token_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        value = value or variables.get_variable("token")
        state.token = value
        return value

    return click.option(
        "-t",
        "--token",
        help="API authorization token",
        type=str,
        expose_value=False,
        callback=callback,
    )(f)


def host_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        value = value or variables.get_variable("host")
        state.host = value
        return value

    return click.option(
        "-H",
        "--host",
        help="Hostname of desired API",
        type=str,
        expose_value=False,
        callback=callback,
    )(f)


def schemes_option(f):
    def callback(ctx, _, value):
        state = ctx.ensure_object(State)
        value = value or variables.get_variable("schemes")
        value = value.split(",") if value is not None else None
        state.schemes = value
        return value

    return click.option(
        "-s",
        "--schemes",
        help="Desired API schemes",
        type=str,
        expose_value=False,
        callback=callback,
    )(f)


def common_options(f):
    f = quiet_option(f)
    f = debug_option(f)
    f = token_option(f)
    f = schemes_option(f)
    f = host_option(f)
    f = hard_option(f)
    return f



from .base import BaseInfoMapper, register
from .utils import int_or_none

CLCAD2 = "CLCAD2"
DP = "DP"


class CLCBioInfoMapper(BaseInfoMapper):
    @classmethod
    def keys(cls):
        return {DP, CLCAD2}

    def do_mapping(self, call_info, allele_position, alt_allele):
        total_depth = call_info.get(DP)

        clcad2 = call_info.get(CLCAD2, [None] * 4)
        ref_depth = int_or_none(clcad2[0])
        alt_depth = int_or_none(clcad2[allele_position])

        return total_depth, alt_depth, ref_depth


register(CLCBioInfoMapper)



import logging
import sys
import structlog


LOGGER_NAME = "gocli"

DEFAULT_PROCESSORS = [
    structlog.stdlib.filter_by_level,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M.%S"),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
]


def setup_logging(quiet=False, verbose=False, json=False):
    # verbose supersedes quiet
    level = logging.WARNING if quiet else logging.INFO
    level = logging.DEBUG if verbose else level

    # root logger
    root = logging.getLogger(LOGGER_NAME)
    root.setLevel(level)
    root.addHandler(logging.StreamHandler(sys.stdout))

    # renderer
    if json:
        renderer = structlog.processors.JSONRenderer()
    else:
        renderer = structlog.dev.ConsoleRenderer()

    # https://structlog.readthedocs.io/en/stable/development.html
    structlog.configure(
        processors=DEFAULT_PROCESSORS + [renderer],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def get_logger(*args, **kwargs):
    return structlog.get_logger(LOGGER_NAME, *args, **kwargs)


def log_with_success(name, success, **kwargs):
    method = get_logger().debug if success else get_logger().error
    msg = "%s %s" % (name, "success" if success else "failed")
    method(msg, **kwargs)



import asyncio
import uuid
from multiprocessing import Process, Queue

from cytoolz.curried import assoc, curry
from glom import glom, Coalesce

from genomoncology import kms
from genomoncology.parse import DocType
from genomoncology.pipeline.transformers import filter_private_keys
from . import Sink

EOL = "EOL"


async def make_call(sdk, data_set, data_set_version, build, in_queue):
    unit = in_queue.get()

    while unit != EOL:
        await kms.annotations.async_load_annotations(
            unit,
            sdk=sdk,
            data_set=data_set,
            data_set_version=data_set_version,
            build=build,
        )
        unit = in_queue.get()


def background_runner(state, data_set, data_set_version, queue):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    sdk = state.create_sdk(async_enabled=True, loop=loop)
    loop.run_until_complete(
        asyncio.gather(
            make_call(sdk, data_set, data_set_version, state.build, queue)
        )
    )


@curry
class LoadAnnotationSink(Sink):
    def __init__(self, _, state, data_set, data_set_version, num_workers=10):
        self.data_set = data_set
        self.data_set_version = data_set_version
        self.queue = Queue(maxsize=1000)
        self.workers = []

        self.num_workers = num_workers

        for _ in range(num_workers):
            worker = Process(
                target=background_runner,
                args=(state, data_set, data_set_version, self.queue),
            )
            worker.start()
            self.workers.append(worker)

    def close(self):
        for _ in range(self.num_workers):
            self.queue.put(EOL)

        for worker in self.workers:
            worker.join()

    def write(self, data):
        data = filter(lambda x: not DocType.HEADER.is_a(x), data)
        data = map(set_uuid(self.data_set, self.data_set_version), data)
        data = list(map(filter_private_keys, data))

        self.queue.put(data)
        return data


UUID = "uuid"


@curry
def set_uuid(data_set, data_set_version, record):
    if UUID not in record:
        uuid_value = glom(record, Coalesce("hgvs_g", "hgvs_c", default=None))

        if uuid_value:
            uuid_value = f"{data_set}|{data_set_version}|{uuid_value}"
            uuid_value = str(uuid.uuid3(uuid.NAMESPACE_DNS, uuid_value))
        else:
            uuid_value = str(uuid.uuid4())

        record = assoc(record, UUID, uuid_value)

    return record



from intervaltree import IntervalTree


# method to map from one chromosome to another (i.e. "M" should map to "MT)
def map_chr(chromosome):
    if chromosome == "M":
        return "MT"
    return chromosome


def clean_chr(chromosome):
    return map_chr(str(chromosome).replace("chr", ""))


class BEDFilter(object):
    def __init__(self, file_path, padding=0):
        self._map = {}

        for line in open(file_path, "r"):
            values = line.strip().split("\t")
            (chromosome, start, end) = values[:3]
            chromosome = clean_chr(chromosome)
            tree = self.get_tree(chromosome)

            start = int(start) - padding
            end = int(end) + 1 + padding  # +1 to make search inclusive

            tree[start:end] = values[3:]

    def __contains__(self, item):
        chromosome, pos = to_chr_pos(item)
        return self.get_tree(chromosome).overlaps(pos)

    def __call__(self, item):
        return item in self

    def get_tree(self, chromosome):
        chromosome = str(chromosome)

        if chromosome not in self._map:
            self._map[chromosome] = IntervalTree()

        return self._map[chromosome]

    def search(self, item):
        chromosome, pos = to_chr_pos(item)
        return self.get_tree(chromosome).search(pos)


# support both plain and chr-prefix versions of 1-22, X, Y, and MT
KNOWN_CHROMOSOMES = list(map(str, range(1, 23))) + ["X", "Y", "MT", "M"]
KNOWN_CHROMOSOMES += list(map(lambda x: f"chr{x}", KNOWN_CHROMOSOMES))


class ValidChromosomeFilter(object):
    def __contains__(self, item):
        chromosome, _ = to_chr_pos(item)
        return chromosome in KNOWN_CHROMOSOMES

    def __call__(self, item):
        return item in self


def to_chr_pos(item):
    if isinstance(item, tuple):
        (chromosome, pos) = (clean_chr(item[0]), item[1])
    else:
        (chromosome, pos) = (clean_chr(item.chrom), item.pos)

    return chromosome, pos



from .base import BaseInfoMapper, register
from .utils import int_or_none

AD = "AD"
RD = "RD"
DP = "DP"


class ADRDDPInfoMapper(BaseInfoMapper):
    @classmethod
    def keys(cls):
        return {DP, AD, RD}

    def do_mapping(self, call_info, allele_position, alt_allele):
        total_depth = call_info.get(DP)
        ref_depth = call_info.get(RD)

        ao = call_info.get(AD)
        alt_depth = (
            int_or_none(ao[allele_position - 1])
            if isinstance(ao, (tuple, list))
            else int_or_none(ao)
        )

        return total_depth, alt_depth, ref_depth


register(ADRDDPInfoMapper)



from .base import Sink, JsonlFileSink, TextFileSink
from .tsv import TsvFileSink
from .excel import ExcelSink
from .alterations import TempAlterationSink
from .vcf import VcfFileSink
from .annotations import LoadAnnotationSink
from .warehouse import LoadWarehouseVariantsSink, LoadWarehouseFeaturesSink


__all__ = (
    "Sink",
    "JsonlFileSink",
    "TextFileSink",
    "TsvFileSink",
    "ExcelSink",
    "TempAlterationSink",
    "VcfFileSink",
    "LoadAnnotationSink",
    "LoadWarehouseVariantsSink",
    "LoadWarehouseFeaturesSink",
)



from cytoolz.curried import curry, unique, concat
from glom import Coalesce

from genomoncology.kms.match import (
    calls_to_alterations,
    transform_match_response,
)
from genomoncology.parse import DocType, __TYPE__
from genomoncology.parse.ensures import ensure_collection


@curry
def match_therapies(calls, diseases=None, sdk=None):
    diseases = ensure_collection(diseases or ["ANY"])

    alterations = calls_to_alterations(calls)

    response = sdk.call_with_retry(
        sdk.therapies.match_therapies_post,
        alterations=alterations,
        diseases=diseases,
    )

    return transform_match_response(response, DEFAULT_SPEC)


DEFAULT_SPEC = {
    __TYPE__: Coalesce(default=DocType.THERAPY.value),
    "drugs": "drugs",
    "response": "response",
    "other_condition": Coalesce("other_condition", default=None),
    "setting": "setting_source_pairs",
    "trigger_alterations": (
        "detected_alterations",
        ["trigger_alterations"],
        lambda x: list(unique(concat(x))),
    ),
    "trigger_diseases": (
        "matched_diseases",
        ["matched_disease"],
        lambda x: list(unique(x)),
    ),
}
